from textwrap import indent
from typing import get_type_hints, Optional, Self, Type, TypeVar, List

from memhax.element import Element
from memhax.utils import CONFIG

T = TypeVar('T')


class StructMember:
    def __init__(self, name: str, parent: type):
        self.name = name
        self.parent = parent
        self._hints = None

    @property
    def hints(self) -> dict:
        if self._hints is None:
            self._hints = get_type_hints(self.parent)
        return self._hints

    @property
    def type(self) -> Type[Element]:
        cls = self.hints[self.name]
        if hasattr(cls, "__orig_class__"):
            cls = cls.__orig_class__
        return cls

    @property
    def size(self) -> int:
        return self.type.sizeof()

    @property
    def alignment(self) -> int:
        return self.type.alignment()


class StructMeta(type):
    def __new__(mcs, name, bases, namespace, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        cls._struct_fields = []

        for parent_class in bases:
            if hasattr(parent_class, '_struct_fields'):
                cls._struct_fields.extend(parent_class._struct_fields)

        for (_name, _) in getattr(cls, '__annotations__', {}).items():
            member = StructMember(_name, cls)
            setattr(cls, _name, member)
            cls._struct_fields.append(member)
        return cls

    def __repr__(self):
        entries = [
            (member.type.typename(member.type), member.name)
            for member in self._struct_fields
        ]
        longest = max(len(typename) for (typename, name) in entries)
        body = "\n".join(
            f"    {typename:<{longest}} {name};"
            for typename, name in entries
        )
        return f"struct {self.__name__} {{\n{body}\n}}"


class Struct(Element[T], metaclass=StructMeta):
    def __init__(self, address: int):
        super().__init__(address)
        self._offset = 0

        # noinspection PyProtectedMember
        for member in self.__class__._struct_fields:
            self._align(member.alignment)
            item = member.type(address + self._offset)
            if isinstance(item, _ForwardRoot):
                item.set_root(self)
            self._offset += member.size

            setattr(self, member.name, item)

    def _align(self, to: int) -> None:
        rem = (self.address + self._offset) % to
        if rem != 0:
            self._offset += to - rem

    @classmethod
    def alignment(cls) -> int:
        return max(item.alignment() for item in get_type_hints(cls).values())

    @classmethod
    def sizeof(cls, instance: Optional[Self] = None) -> int:
        offset = 0
        for member in cls._struct_fields:
            _type = member.type
            align = _type.alignment()
            rem = offset % align
            if rem != 0:
                offset += align - rem
            offset += _type.sizeof(getattr(instance, member.name, instance))
        return offset

    @classmethod
    def offsetof(cls, field) -> int:
        offset = 0
        for member in cls._struct_fields:
            if member.name == field.name:
                return offset
            rem = offset % member.alignment
            if rem != 0:
                offset += member.alignment - rem
            offset += member.size
        raise ValueError(f'Field {field.name} not found in {cls.__name__}')

    @classmethod
    def offset_after(cls, field) -> int:
        return cls.offsetof(field) + field.size

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return cls.__name__

    def repr_rich(self, visited: List[int]) -> str:
        attrs = [(item.name, getattr(self, item.name)) for item in self.__class__._struct_fields]
        if CONFIG["pretty_repr"]:
            body = "\n" + ",\n".join(indent(f"{attr}={field.repr_handler(visited[:])}", "    ") for attr, field in attrs) + "\n"
        else:
            body = ", ".join(f"{attr}={field.repr_handler(visited[:])}" for attr, field in attrs)
        return f"{self.__class__.__name__}({body})"


class _ForwardRoot:
    def __init__(self, *args, **kwargs):
        self.root = None

    def set_root(self, root: 'Struct'):
        self.root = root



import atexit
from contextlib import contextmanager
from typing import get_args, Any, Optional, Tuple


_MEMORY = open("/proc/self/mem", "r+b")
CONFIG = {
    "simplified_repr": True,
    "pretty_repr": True,
    "hide_pointers_repr": True,
}
__all__ = ("CONFIG", "memory", "instance_type_args")


def instance_type_args(instance: Any) -> Optional[Tuple[Any, ...]]:
    _orig = getattr(instance, "__orig_class__", None)
    if _orig is None:
        return None
    return get_args(_orig)


@contextmanager
def memory(address: Optional[int] = None):
    tmp = _MEMORY.tell()
    if address is not None:
        _MEMORY.seek(address)
    yield _MEMORY
    _MEMORY.seek(tmp)


@atexit.register
def __close_memory():
    _MEMORY.close()



from memhax.native.native import Integer, Py_ssize_t
from memhax.native.native_complex import Pointer, NullTerminatedString
from memhax.native.structs import Struct


class PyAsyncMethods(Struct[None]):
    am_await: Pointer
    am_aiter: Pointer
    am_anext: Pointer
    am_send: Pointer


class PyNumberMethods(Struct[None]):
    nb_add: Pointer
    nb_subtract: Pointer
    nb_multiply: Pointer
    nb_remainder: Pointer
    nb_divmod: Pointer
    nb_power: Pointer
    nb_negative: Pointer
    nb_positive: Pointer
    nb_absolute: Pointer
    nb_bool: Pointer
    nb_invert: Pointer
    nb_lshift: Pointer
    nb_rshift: Pointer
    nb_and: Pointer
    nb_xor: Pointer
    nb_or: Pointer
    nb_int: Pointer
    nb_reserved: Pointer
    nb_float: Pointer
    nb_inplace_add: Pointer
    nb_inplace_subtract: Pointer
    nb_inplace_multiply: Pointer
    nb_inplace_remainder: Pointer
    nb_inplace_power: Pointer
    nb_inplace_lshift: Pointer
    nb_inplace_rshift: Pointer
    nb_inplace_and: Pointer
    nb_inplace_xor: Pointer
    nb_inplace_or: Pointer
    nb_floor_divide: Pointer
    nb_true_divide: Pointer
    nb_inplace_floor_divide: Pointer
    nb_inplace_true_divide: Pointer
    nb_index: Pointer
    nb_matrix_multiply: Pointer
    nb_inplace_matrix_multiply: Pointer


class PySequenceMethods(Struct[None]):
    sq_length: Pointer
    sq_concat: Pointer
    sq_repeat: Pointer
    sq_item: Pointer
    was_sq_slice: Pointer
    sq_ass_item: Pointer
    was_sq_ass_slice: Pointer
    sq_contains: Pointer
    sq_inplace_concat: Pointer
    sq_inplace_repeat: Pointer


class PyMappingMethods(Struct[None]):
    mp_length: Pointer
    mp_subscript: Pointer
    mp_ass_subscript: Pointer


class PyBufferProcs(Struct[None]):
    bf_getbuffer: Pointer
    bf_releasebuffer: Pointer


class PyMethodDef(Struct[None]):
    ml_name: Pointer[NullTerminatedString]
    ml_meth: Pointer
    ml_flags: Integer
    ml_doc: Pointer[NullTerminatedString]


class PyMemberDef(Struct[None]):
    name: Pointer[NullTerminatedString]
    type: Integer
    offset: Py_ssize_t
    flags: Integer
    doc: Pointer[NullTerminatedString]


class PyGetSetDef(Struct[None]):
    name: Pointer[NullTerminatedString]
    get: Pointer
    set: Pointer
    doc: Pointer[NullTerminatedString]
    closure: Pointer



from typing import Self, Optional, Type

from memhax.element import PackedElement


class LongLong(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "q"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "long long"


class UnsignedLongLong(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "Q"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "unsigned long long"


class Long(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "l"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "long"


class UnsignedLong(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "L"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "unsigned long"


class Integer(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "i"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "int"


class UnsignedInteger(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "I"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "unsigned int"


class Short(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "h"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "short"


class UnsignedShort(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "H"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "unsigned short"


class Byte(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "b"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "byte"


class UnsignedByte(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "B"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "unsigned byte"


class Float(PackedElement[float]):
    @classmethod
    def fmt(cls) -> str:
        return "f"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "float"


class Double(PackedElement[float]):
    @classmethod
    def fmt(cls) -> str:
        return "d"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "double"


class Char(PackedElement[bytes]):
    @classmethod
    def fmt(cls) -> str:
        return "c"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "char"


class Boolean(PackedElement[bool]):
    @classmethod
    def fmt(cls) -> str:
        return "?"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "bool"


class Py_ssize_t(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "n"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "Py_ssize_t"


class Py_size_t(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "N"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "Py_size_t"



from ctypes import sizeof, c_wchar

from memhax.utils import memory
from memhax.cpython.object import PyVarObject, PyObject
from memhax.native.native import Double, Py_ssize_t, Char
from memhax.native.native_complex import PropertySizeArray, Pointer
from memhax.native.native_size import uint32_t
from memhax.native.structs import Struct


class PyLongObject(PyVarObject, Struct[int]):
    ob_digit: PropertySizeArray[uint32_t, lambda self: abs(self.ob_size())]


class PyFloatObject(PyVarObject, Struct[float]):
    ob_fval: Double


class PyBytesObject(PyVarObject, Struct[bytes]):
    ob_shash: Py_ssize_t
    ob_sval: PropertySizeArray[Char, lambda self: self.ob_size()]


class PyASCIIObject(PyObject, Struct[str]):
    length: Py_ssize_t
    hash: Py_ssize_t
    state: uint32_t
    wstr: Pointer

    def _string(self, interned: int, kind: int, compact: bool, _ascii: bool, ready: bool) -> str:
        if not (kind == 1 and compact and _ascii and ready):
            raise ValueError("Not a PyASCIIObject valid string:", interned, kind, compact, _ascii, ready)
        addr = self.address + PyASCIIObject.offset_after(self.wstr)
        with memory(addr) as mem:
            data = mem.read(self.length())
            return data.decode("ascii")

    def value(self) -> str:
        state = self.state()
        interned = state & 0b11
        kind = (state >> 2) & 0b111
        compact = (state >> 5) & 0b1
        _ascii = (state >> 6) & 0b1
        ready = (state >> 7) & 0b1
        return self._string(interned, kind, compact, _ascii, ready)


class PyCompactUnicodeObject(PyASCIIObject):
    utf8_length: Py_ssize_t
    utf8: Pointer
    wstr_length: Py_ssize_t

    def _string(self, interned: int, kind: int, compact: bool, _ascii: bool, ready: bool) -> str:
        if kind == 1 and compact and _ascii and ready:
            return super()._string(interned, kind, compact, _ascii, ready)
        if kind != 0 and compact and not _ascii and ready:
            addr = self.address + PyCompactUnicodeObject.offset_after(PyCompactUnicodeObject.wstr_length)
            with memory(addr) as mem:
                data = mem.read(self.length() * kind)
                fmt = {
                    1: "utf-8",
                    2: "utf-16",
                    4: "utf-32"
                }
                return data.decode(fmt[kind])
        raise ValueError("Not a PyCompactUnicodeObject valid string:", interned, kind, compact, _ascii, ready)


class PyUnicodeObject(PyCompactUnicodeObject):
    data: Pointer

    def _string(self, interned: int, kind: int, compact: bool, _ascii: bool, ready: bool) -> str:
        if kind == 1 and compact and _ascii and ready:
            return super()._string(interned, kind, compact, _ascii, ready)  # PyASCIIObject
        if kind != 0 and compact and not _ascii and ready:
            return super()._string(interned, kind, compact, _ascii, ready)  # PyCompactUnicodeObject
        if kind == 0 and not compact and not _ascii and not ready:
            with memory(self.wstr()) as mem:
                data = mem.read(self.wstr_length() * sizeof(c_wchar))  # Depends on ctypes :(
                fmt = {
                    2: "utf-16",
                    4: "utf-32"
                }
                return data.decode(fmt[sizeof(c_wchar)])
        elif kind != 0 and not compact and ready:
            with memory(self.data()) as mem:
                data = mem.read(self.length() * kind)
                fmt = {
                    1: "utf-8",
                    2: "utf-16",
                    4: "utf-32"
                }
                return data.decode(fmt[kind])
        raise ValueError("Not a PyUnicodeObject valid string:", interned, kind, compact, _ascii, ready)



from __future__ import annotations

from struct import calcsize, unpack, pack
from typing import TypeVar, Self, Union, Optional, Generic, get_type_hints, Type, List

from memhax.utils import memory, CONFIG

T = TypeVar('T')


class Element(Generic[T]):
    def __init__(self, address: int):
        self.address = address

    def __call__(self, instance: Optional[Union[Self, T]] = None) -> T:
        if instance is None:
            return self.get()
        else:
            return self.set(instance)

    def get(self) -> T:
        raise NotImplementedError(self.__class__.__name__)

    def set(self, value: Union[Self, T]) -> T:
        raise NotImplementedError(self.__class__.__name__)

    @classmethod
    def alignment(cls) -> int:
        raise NotImplementedError(cls.__name__)

    @classmethod
    def sizeof(cls, instance: Optional[Self] = None) -> int:
        raise NotImplementedError(cls.__name__)

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        raise NotImplementedError(cls.__name__)

    def repr_handler(self, visited: List[int]):
        visited.append(self.address)
        if CONFIG["simplified_repr"]:
            return self.repr_simple(visited[:])
        else:
            return self.repr_rich(visited[:])

    def repr_simple(self, visited: List[int]) -> str:
        return self.repr_rich(visited[:])

    def repr_rich(self, visited: List[int]) -> str:
        return super().__repr__()

    def __repr__(self):
        return self.repr_handler([])


class PackedElement(Element[T]):
    @classmethod
    def fmt(cls) -> str:
        raise NotImplementedError(cls.__name__)

    @classmethod
    def alignment(cls) -> int:
        return calcsize("@"+cls.fmt())

    @classmethod
    def sizeof(cls, instance: Optional[Self] = None) -> int:
        return calcsize("@"+cls.fmt())

    def get(self) -> T:
        with memory(self.address) as mem:
            item = unpack("@"+self.fmt(), mem.read(self.sizeof()))
        return item[0] if len(item) == 1 else item

    def set(self, value: Union[Self, T]) -> T:
        with memory(self.address) as mem:
            if self.__class__ == value.__class__:
                value = value()
            if isinstance(value, tuple):
                mem.write(pack("@"+self.fmt(), *value))
            else:
                mem.write(pack("@"+self.fmt(), value))
        return value

    def repr_simple(self, visited: List[int]) -> str:
        return repr(self.get())

    def repr_rich(self, visited: List[int]) -> str:
        return f"{self.__class__.__name__}({self.get()})"



from textwrap import indent
from typing import TypeVar, get_args, Generic, List, Optional, Self, Type

from memhax.native.structs import _ForwardRoot
from memhax.utils import memory, CONFIG
from memhax.element import PackedElement, Element
from memhax.utils import instance_type_args

E = TypeVar('E', bound=Element)
T = TypeVar('T')
S = TypeVar('S')


class RawPointer(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return "P"


class Pointer(PackedElement[E], _ForwardRoot):
    raw: RawPointer

    def __init__(self, address: int):
        super().__init__(address)
        self.raw = RawPointer(self.address)

    @classmethod
    def fmt(cls) -> str:
        return "P"

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        if complete_type is None or isinstance(complete_type, type):
            return "void*"
        _args = get_args(complete_type)
        _type = _args[0]
        return f"{_type.typename(_type)}*"

    def get(self) -> E:
        _args = instance_type_args(self)
        if _args is None:
            raise TypeError("Cannot get value of a void pointer")
        _type = _args[0]
        item = _type(self.raw())
        if isinstance(item, _ForwardRoot):
            item.set_root(self.root)
        return item

    def set(self, value: E) -> None:
        super().set(value.address)

    def repr_handler(self, visited: List[int]):
        _args = instance_type_args(self)

        if self.raw() == 0:
            return "NULL"
        if _args is None:
            return f"Pointer(0x{self.raw():X})"
        return super().repr_handler(visited)

    def repr_simple(self, visited: List[int]) -> str:
        child = self.get()
        if child.address in visited:
            child_repr = "..."
        else:
            child_repr = child.repr_handler(visited[:])
        if CONFIG["hide_pointers_repr"]:
            return child_repr
        else:
            return f"*{child_repr}"

    def repr_rich(self, visited: List[int]) -> str:
        child = self.get()
        if child.address in visited:
            child_repr = "..."
        else:
            child_repr = child.repr_handler(visited[:])
        return f"Pointer({child_repr})"


class _ArrayCommon(Element[List[E]], _ForwardRoot):
    @classmethod
    def alignment(cls) -> int:
        return 1

    @classmethod
    def sizeof(cls, instance: Optional[Self] = None) -> int:
        return 0

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        _args = get_args(complete_type)
        _type = _args[0]
        _size = _args[1] if len(_args) > 1 else ""
        if callable(_size):
            _size = ""

        return f"{_type.typename(_type)}[{_size}]"

    def set(self, value: T) -> None:
        original = self.get()
        if len(value) != len(original):
            raise ValueError("Cannot set array of different size")

        for old, new in zip(original, value):
            old(new)

    def __getitem__(self, item):
        return self.get()[item]

    def __setitem__(self, key, value):
        self.get()[key](value)

    def repr_simple(self, visited: List[int]) -> str:
        items = self.get()
        if CONFIG["pretty_repr"]:
            body = "\n" + ",\n".join(indent(item.repr_handler(visited[:]), "    ") for item in items) + "\n"
        else:
            body = ", ".join(item.repr_handler(visited[:]) for item in items)
        return f"[{body}]"

    def repr_rich(self, visited: List[int]) -> str:
        return f"{self.__class__.__name__}({self.repr_simple(visited)})"

    def set_root(self, root: 'Struct'):
        super().set_root(root)
        for child in self.get():
            if isinstance(child, _ForwardRoot):
                child.set_root(root)


class StaticSizeArray(_ArrayCommon[T], Generic[T, S]):
    def get(self) -> T:
        _type, _size = instance_type_args(self)
        items = []
        address = self.address

        _align = _type.alignment()
        _itemsize = _type.sizeof()

        for _ in range(_size):
            rem = (address % _align)
            if rem != 0:
                address += _align - rem

            items.append(_type(address))
            address += _itemsize
        return items


class PropertySizeArray(_ArrayCommon[T], Generic[T, S]):
    def get(self) -> T:
        _type, _size = instance_type_args(self)
        items = []
        address = self.address

        _align = _type.alignment()
        _itemsize = _type.sizeof()

        for _ in range(_size(self.root)):
            rem = (address % _align)
            if rem != 0:
                address += _align - rem

            items.append(_type(address))
            address += _itemsize
        return items


class NullTerminatedArray(_ArrayCommon[T], Generic[T]):
    def get(self) -> T:
        _type = instance_type_args(self)[0]
        items = []
        address = self.address

        _align = _type.alignment()
        _itemsize = _type.sizeof()

        while True:
            rem = (address % _align)
            if rem != 0:
                address += _align - rem

            with memory(address) as mem:
                if all(b == 0 for b in mem.read(_itemsize)):
                    return items

            items.append(_type(address))
            address += _itemsize


class NullTerminatedString(Element):
    def get(self) -> T:
        chars = []
        address = self.address
        while True:
            with memory(address) as mem:
                if (byte := mem.read(1)) == b"\x00":
                    return "".join(chars)

            chars.append(byte.decode())
            address += 1

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return "char[]"

    @classmethod
    def alignment(cls) -> int:
        return 1

    @classmethod
    def sizeof(cls, instance: Optional[Self] = None) -> int:
        if instance is None:
            return 0
        return len(instance.get()) + 1

    def repr_simple(self, visited: List[int]) -> str:
        if self.address < 1000:
            return "INVALID_STRING"
        return repr(self.get())

    def repr_rich(self, visited: List[int]) -> str:
        return f"NullTerminatedString({self.repr_simple(visited)})"



from memhax.cpython.object import PyVarObject, PyObject
from memhax.native.native import Py_ssize_t
from memhax.native.native_complex import PropertySizeArray, Pointer, StaticSizeArray
from memhax.native.native_size import uint64_t
from memhax.native.structs import Struct


class PyTupleObject(PyVarObject, Struct[tuple]):
    ob_item: PropertySizeArray[Pointer[PyObject], lambda self: self.ob_size()]  # how am I gonna pass `self` [PyTupleObject] to this function


class PyListObject(PyVarObject, Struct[list]):
    ob_item: Pointer[PropertySizeArray[Pointer[PyObject], lambda self: self.allocated()]]
    allocated: Py_ssize_t


class _PySetEntry(Struct[None]):
    key: Pointer[PyObject]
    hash: Py_ssize_t


class PySetObject(PyObject, Struct[set]):
    fill: Py_ssize_t
    used: Py_ssize_t
    mask: Py_ssize_t
    table: Pointer[PropertySizeArray[_PySetEntry, lambda self: self.fill()]]
    hash: Py_ssize_t
    finger: Py_ssize_t
    smalltable: StaticSizeArray[_PySetEntry, 8]
    weakreflist: Pointer[PyObject]


class PyDictObject(PyObject, Struct[dict]):
    ma_used: Py_ssize_t
    ma_version_tag: uint64_t
    # TODO: PyDictKeysObject; It's a fairly complex struct so not implemented yet
    ma_keys: Pointer
    ma_values: Pointer[PropertySizeArray[Pointer[PyObject], lambda self: self.ma_used()]]



from __future__ import annotations

from typing import List

from memhax.cpython.type_attributes import PyAsyncMethods, PyNumberMethods, PySequenceMethods, PyMappingMethods, PyBufferProcs, PyMethodDef, PyMemberDef, \
    PyGetSetDef
from memhax.native.structs import Struct
from memhax.native.native import Py_ssize_t, UnsignedInteger
from memhax.native.native_complex import Pointer, NullTerminatedArray, NullTerminatedString


class PyObject(Struct[object]):
    ob_refcnt: Py_ssize_t
    ob_type: Pointer[PyTypeObject]

    def incref(self):
        self.ob_refcnt(self.ob_refcnt() + 1)

    def decref(self):
        self.ob_refcnt(self.ob_refcnt() - 1)

    def _reinterpret(self) -> object:
        from memhax.cpython.collections import PyTupleObject
        tup = (None,)
        _tuple = PyTupleObject(id(tup))
        _tuple.ob_item[0]().decref()
        _tuple.ob_item[0].raw(self.address)
        self.incref()
        return tup[0]

    def get(self) -> object:
        return self._reinterpret()


class PyVarObject(PyObject):
    ob_size: Py_ssize_t


class PyTypeObject(PyVarObject, Struct[type]):
    tp_name: Pointer[NullTerminatedString]
    tp_basicsize: Py_ssize_t
    tp_itemsize: Py_ssize_t
    tp_dealloc: Pointer
    tp_vectorcall_offset: Py_ssize_t
    tp_getattr: Pointer
    tp_setattr: Pointer
    tp_as_async: Pointer[PyAsyncMethods]
    tp_repr: Pointer
    tp_as_number: Pointer[PyNumberMethods]
    tp_as_sequence: Pointer[PySequenceMethods]
    tp_as_mapping: Pointer[PyMappingMethods]
    tp_hash: Pointer
    tp_call: Pointer
    tp_str: Pointer
    tp_getattro: Pointer
    tp_setattro: Pointer
    tp_as_buffer: Pointer[PyBufferProcs]
    tp_flags: Py_ssize_t
    tp_doc: Pointer[NullTerminatedString]
    tp_traverse: Pointer
    tp_clear: Pointer
    tp_richcompare: Pointer
    tp_weaklistoffset: Py_ssize_t
    tp_iter: Pointer
    tp_iternext: Pointer
    tp_methods: Pointer[NullTerminatedArray[PyMethodDef]]
    tp_members: Pointer[NullTerminatedArray[PyMemberDef]]
    tp_getset: Pointer[NullTerminatedArray[PyGetSetDef]]
    tp_base: Pointer[PyTypeObject]
    tp_dict: Pointer[PyObject]
    tp_descr_get: Pointer
    tp_descr_set: Pointer
    tp_dictoffset: Py_ssize_t
    tp_init: Pointer
    tp_alloc: Pointer
    tp_new: Pointer
    tp_free: Pointer
    tp_is_gc: Pointer
    tp_bases: Pointer[PyObject]
    tp_mro: Pointer[PyObject]
    tp_cache: Pointer
    tp_subclasses: Pointer
    tp_weaklist: Pointer
    tp_del: Pointer
    tp_version_tag: UnsignedInteger
    tp_finalize: Pointer
    tp_vectorcall: Pointer

    def repr_simple(self, visited: List[int]) -> str:
        return repr(self.get())



import sys
from typing import Optional, Type, Self, Union

from memhax.element import PackedElement


class uintX_t(PackedElement[int]):
    @classmethod
    def fmt(cls) -> str:
        return f"{cls.num_bytes()}B"

    def get(self) -> int:
        _bytes = super().get()
        _size = self.num_bytes()
        if sys.byteorder == "big":
            _bytes = _bytes[::-1]
        return sum(_bytes[i] << (8 * i) for i in range(_size))

    def set(self, value: Union[Self, int]) -> int:
        if not isinstance(value, int):
            value = value()
        _bytes = [
            (value >> n) & 0xFF
            for n in range(self.num_bytes())
        ]
        if sys.byteorder == "big":
            _bytes = _bytes[::-1]
        super().set(tuple(_bytes))
        return value

    @classmethod
    def typename(cls, complete_type: Optional[Type[Self]] = None) -> str:
        return f"uint{cls.num_bytes() * 8}_t"

    @classmethod
    def num_bytes(cls) -> int:
        raise NotImplementedError


class uint8_t(uintX_t):
    @classmethod
    def num_bytes(cls) -> int:
        return 1


class uint16_t(uintX_t):
    @classmethod
    def num_bytes(cls) -> int:
        return 2


class uint32_t(uintX_t):
    @classmethod
    def num_bytes(cls) -> int:
        return 4


class uint64_t(uintX_t):
    @classmethod
    def num_bytes(cls) -> int:
        return 8



from setuptools import setup

with open("README.md", "r") as fp:
    long_description = fp.read()

try:
    file = open("/proc/self/mem", "r+b", buffering=0)
    file.close()
except FileNotFoundError:
    raise RuntimeError("This package requires access to /proc/self/mem, which does not exist on this system.")

setup(
    name="memhax",
    version="0.1.0",
    packages=['memhax', 'memhax.cpython', 'memhax.native'],
    description="A Python library for getting access to raw memory and internals",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.11,<3.12',  # TODO: Support other python versions
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Operating System :: Unix",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Typing :: Typed",
    ],
)
